/* Styling for the E-lands logo and the title of the report */
.header_title_logo {
    text-align: center;
    color: #7D4E57;
    border-style: solid;
    border-width: thin;
    font-weight: bold;
    width: 91.5%;
    }

/* Creates containers around some headings */
h2 {
    text-align: center;
    borderStyle: solid;
    borderWidth: thin;
    padding-left: 1rem;
    font-size: 250%;
    width: 100%;
    margin-left: auto;
    margin-right: auto;
    background: #085237;
    color: #FFFFFF;
    verticalAlign: middle;
    }

h3 {
    text-align: left;
    margin-top: 20px;
    font-weight: bold;
    color: #8c3604;
    }

h4 {
    border-left: 20px solid #8c3604;
    background: #ffffff;
    padding-top: 1px;
    padding-left: 5px;
    padding-bottom: 1px;
    color: #048BA8;
    font-weight: bold;
    margin-top: 20px;
    }

/* styling properties for the images and graphs */
.graphs_ts {
    margin-left: auto;
    margin-right: auto;
    display: block;
    margin-bottom: 2cm;
  }

/* Styling for important information under the title, such as version, date, etc. */
.imp_info {
   text-align: right;
   padding: 5px;
   }

/* Styling for important information like project and scenario names */
.imp_info2 {
   text-align: left;
   }

@media screen {

    /* Styling for the E-lands logo and the title of the report */
    .header_title_logo {
        margin: 30px;
}
    /* Styling for the log messages */
    .list-log {
        text-align: left;
        font-size: 20pt;
}
    .tableplay {
        font-size: 12pt;
        margin-top:15px;
        margin-bottom: 2cm;
    }

    /* Styling for the pre-titles of the graphs */
    .graph__pre-title {
        text-align: left;
        fontSize: 40px;
        position: relative;
        left: 0;
        height: 20%;
        margin-left: 30px;
        borderLeft: 20px solid #8c3604;
        background: #ffffff;
        paddingTop: 1px;
        paddingBottom: 1px;
        paddingLeft: 30px;
        paddingRight: 60px;
        }

    h1 {
        font-size: 320%;
        font-weight: bold;
        }

    /* Styles for subheadings in the report */
    .projdataheading {
        position: relative;
        left: 0;
        height: 20%;
        margin: 30px;
        border-left: 20px solid #8c3604;
        background: #ffffff;
        paddingTop: 1px;
        paddingBottom: 1px;
        paddingLeft: 30px;
        paddingRight: 60px;
        }

    /* Styling for important information under the title, such as version, date, etc. */
    .imp_info {
        font-size: 10pt;
        margin: 30px;
        }


    /* Styling for important information like project and scenario names */
    .imp_info2 {
        font-size: 20pt;
        margin: 30px;
        }

    /* Styling for blocks of texts or paragraphs  */
    .blockoftext {
        text-align: left;
        font-size: 20pt;
        margin: 30px;
        }


    h3 {
        margin-left: 30px;
        font-size: 30pt;
        }

    h4 {
        margin-left: 30px;
        font-size: 23pt;
        }

    .report_table_title {
        color: #048BA8;
        font-weight: bold;
        margin-top: 20px
    }

    /* Styling for the demand data section of the report */
    .demandmatter {
        text-align: left;
        fontSize: 40px;
        margin: 30px;
        }

    /* Provides style for headings of each segment of the report */
    .heading1 {
        text-align: left;
        margin: 30px;
        fontSize: 60px;
        color: #8c3604;
        }

    /* Creates a heading underline */
    .horizontal_line {
        position:relative;
        border-bottom:1px solid black;
        padding-bottom:3px;
        width: 100%;
        margin-left:auto;
        margin-right:auto;
        }

    /* Styling for the section containing the project data */
    .projdata {
        width: 48%;
        margin: 30px;
        fontSize: 40px;
        }

    p {
       font-size: 140%;
       margin:30px;
        }

    iframe {
        border: 1px solid black;
        width: 50%;
        display:block;
    }


/* on-screen styles */


  .staticimagepdf {
    display: none;
  }

  .print-only {
    display: none;
  }
}

/* print styles */
@media print {

  iframe {
    display: none;
   }

  /* */
  #location-map-div {

    }

  .no-print > div {
    display: none;
  }

  .list-log{
    text-align: left;
    font-size: 15pt;
  }

  .dash-plots {
  display: block;
  margin-left: auto;
  margin-right: auto;
  width: 70%;
  border: 20px dotted blue;
  }

  .location-map-column {
    width: 48%;
    // padding-left: 2%;
    // padding-right: 2%;
    // display: inline-block;
    }

  .tableplay {
    page-break-inside:avoid;
    font-size: 9pt;
    margin-top:1cm;
    margin-bottom: 1cm;
    margin-left:1cm;
    margin-right:1cm;

  }

  h3 {
    font-size: 26pt;
    margin-bottom: 0px;
    }

  .horizontal_line {
    display: block;
    height: 1px;
    border: 0;
    border-top: 1px solid #a3a3a3;
    margin-top: 2mm;
    padding: 0;
    }

  h4 {
    font-size: 23pt;
    }

  .report_table{
    max-width: 100%;
  }

  table{ width: 100% }

   /* Styling for blocks of texts or paragraphs  */
   .blockoftext {
    text-align: justify;
    text-justify: inter-word;
    font-size: 15pt;
   }

   /* Styling for important information under the title, such as version, date, etc. */
   .imp_info {
       font-size: 12pt;
    }

    /* Styling for important information like project and scenario names */
   .imp_info2 {
       font-size: 15pt;
       margin-bottom: 0.5cm;
   }

   .header_title_logo {
        padding-top: 0.5cm;
   }

  section{
    display: block;
    width: 100%
  }

  .grid-x{
    display: block;
  }

  .align-center{
    align: center;
  }

  .cell{
    display: block;
    width: 100%;
  }
  .small-10{
    width: 100%;
  }
  .small_offset-1{
    width: 100%;
  }

header, footer, aside, nav, form, iframe, .menu, .hero, .adslot {
  display: none;
}

}

/* set margins of the document */
@page {
  margin-top: 1cm;
  margin-bottom: 1cm;
  margin-left: 2.25cm;
  margin-right: 2.25cm;
}
